@@ -29,6 +29,7 @@
 #include "boxed.h"
 #include "union.h"
 #include <gjs/gjs.h>
+#include <gjs/compat.h>
 
 #include <util/log.h>
 
@@ -41,39 +42,48 @@
 #include <errno.h>
 #include <string.h>
 
+/* We use guint8 for arguments; functions can't
+ * have more than this.
+ */
+#define GJS_ARG_INDEX_INVALID G_MAXUINT8
+
 typedef struct {
     GIFunctionInfo *info;
-    guint n_destroy_notifies;
-    guint n_user_data;
-    guint expected_js_argc;
-    guint js_out_argc;
-    guint inout_argc;
+
+    /* We only support at most one of each of these */
+    guint8 callback_index;
+    guint8 destroy_notify_index;
+    guint8 user_data_index;
+
+    guint8 expected_js_argc;
+    guint8 js_out_argc;
+    guint8 inout_argc;
     GIFunctionInvoker invoker;
 } Function;
 
 static struct JSClass gjs_function_class;
 
-typedef struct {
-    GSList *arguments; /* List of jsval, that need to be unrooted after call*/
-    GSList *invoke_infos; /* gjs_callback_invoke_info_free for each*/
-    gint arg_index;
-} GjsCallbackInfo;
+/* Because we can't free the mmap'd data for a callback
+ * while it's in use, this list keeps track of ones that
+ * will be freed the next time we invoke a C function.
+ */
+static GSList *completed_trampolines = NULL;  /* GjsCallbackTrampoline */
+
+static gboolean trampoline_globals_initialized = FALSE;
+static struct {
+    GICallableInfo *info;
+    ffi_cif cif;
+    ffi_closure *closure;
+} global_destroy_trampoline;
 
 typedef struct {
     JSContext *context;
     GICallableInfo *info;
-    jsval function;
-    ffi_cif *cif;
-    gboolean need_free_arg_types;
+    jsval js_function;
+    ffi_cif cif;
     ffi_closure *closure;
-    GjsCallbackInfo callback_info;
-/* We can not munmap closure in ffi_callback.
- * This memory page will be used after it.
- * So we need to know when GjsCallbackInvokeInfo becoming unneeded.*/
-    gboolean in_use;
-} GjsCallbackInvokeInfo;
-
-static void gjs_callback_invoke_info_free_content(gpointer user_data);
+    GIScopeType scope;
+} GjsCallbackTrampoline;
 
 GJS_DEFINE_PRIV_FROM_JS(Function, gjs_function_class)
 
@@ -93,7 +103,7 @@ GJS_DEFINE_PRIV_FROM_JS(Function, gjs_function_class)
 static JSBool
 function_new_resolve(JSContext *context,
                      JSObject  *obj,
-                     jsval      id,
+                     jsid       id,
                      uintN      flags,
                      JSObject **objp)
 {
@@ -102,7 +112,7 @@ function_new_resolve(JSContext *context,
 
     *objp = NULL;
 
-    if (!gjs_get_string_id(id, &name))
+    if (!gjs_get_string_id(context, id, &name))
         return JS_TRUE; /* not resolved, but no error */
 
     priv = priv_from_js(context, obj);
@@ -116,42 +126,15 @@ function_new_resolve(JSContext *context,
 }
 
 static void
-gjs_callback_invoke_info_free(gpointer data)
+gjs_callback_trampoline_free(GjsCallbackTrampoline *trampoline)
 {
-    GjsCallbackInvokeInfo *invoke_info = (GjsCallbackInvokeInfo *)data;
-
-    gjs_callback_invoke_info_free_content(invoke_info);
-    if (invoke_info->info)
-        g_base_info_unref((GIBaseInfo*)invoke_info->info);
-    munmap(invoke_info->closure, sizeof(ffi_closure));
-    if (invoke_info->need_free_arg_types)
-        g_free(invoke_info->cif->arg_types);
-    g_slice_free(ffi_cif, invoke_info->cif);
-
-    g_slice_free(GjsCallbackInvokeInfo, data);
-}
-
-static void
-gjs_callback_invoke_info_free_content(gpointer user_data)
-{
-    GSList *l;
-    GjsCallbackInvokeInfo *invoke_info = (GjsCallbackInvokeInfo*)user_data;
-
-    for (l = invoke_info->callback_info.arguments; l; l = g_slist_next(l)) {
-        jsval *val = l->data;
-        JS_RemoveRoot(invoke_info->context, val);
-        g_slice_free(jsval, val);
-    }
-    g_slist_free(invoke_info->callback_info.arguments);
-    invoke_info->callback_info.arguments = NULL;
-
-    g_slist_foreach(invoke_info->callback_info.invoke_infos, (GFunc)gjs_callback_invoke_info_free, NULL);
-    g_slist_free(invoke_info->callback_info.invoke_infos);
-    invoke_info->callback_info.invoke_infos = NULL;
+    JS_RemoveValueRoot(trampoline->context, &trampoline->js_function);
+    g_callable_info_free_closure(trampoline->info, trampoline->closure);
+    g_base_info_unref( (GIBaseInfo*) trampoline->info);
+    g_slice_free(GjsCallbackTrampoline, trampoline);
 }
 
 /* This is our main entry point for ffi_closure callbacks.
- * (there's a similar one for GDestroyNotify callbacks below)
  * ffi_prep_closure is doing pure magic and replaces the original
  * function call with this one which gives us the ffi arguments,
  * a place to store the return value and our use data.
@@ -164,51 +147,50 @@ gjs_callback_closure(ffi_cif *cif,
                      void **args,
                      void *data)
 {
-    GjsCallbackInvokeInfo *invoke_info = (GjsCallbackInvokeInfo *) data;
-    int i, n_args;
+    GjsCallbackTrampoline *trampoline;
+    int i, n_args, n_jsargs;
     jsval *jsargs, rval;
     GITypeInfo ret_type;
+    gboolean success = FALSE;
 
-    g_assert(invoke_info != NULL);
+    trampoline = data;
+    g_assert(trampoline);
 
-    n_args = g_callable_info_get_n_args(invoke_info->info);
+    n_args = g_callable_info_get_n_args(trampoline->info);
 
     g_assert(n_args >= 0);
 
-    jsargs = (jsval*)g_new0(jsval, n_args);
-    for (i = 0; i < n_args; i++) {
+    jsargs = (jsval*)g_newa(jsval, n_args);
+    for (i = 0, n_jsargs = 0; i < n_args; i++) {
         GIArgInfo arg_info;
         GITypeInfo type_info;
 
-        g_callable_info_load_arg(invoke_info->info, i, &arg_info);
+        g_callable_info_load_arg(trampoline->info, i, &arg_info);
         g_arg_info_load_type(&arg_info, &type_info);
 
-        if (g_type_info_get_tag(&type_info) == GI_TYPE_TAG_VOID) {
-            jsargs[i] = *((jsval*)args[i]);
-        } else if (!gjs_value_from_g_argument(invoke_info->context,
-                                              &jsargs[i],
-                                              &type_info,
-                                              args[i])) {
-            gjs_throw(invoke_info->context, "could not convert argument of callback");
-            g_free(jsargs);
-            return;
-        }
+        /* Skip void * arguments */
+        if (g_type_info_get_tag(&type_info) == GI_TYPE_TAG_VOID)
+            continue;
+
+        if (!gjs_value_from_g_argument(trampoline->context,
+                                       &jsargs[n_jsargs++],
+                                       &type_info,
+                                       args[i]))
+            goto out;
     }
 
-    if (!JS_CallFunctionValue(invoke_info->context,
+    if (!JS_CallFunctionValue(trampoline->context,
                               NULL,
-                              invoke_info->function,
-                              n_args,
+                              trampoline->js_function,
+                              n_jsargs,
                               jsargs,
                               &rval)) {
-        gjs_throw(invoke_info->context, "Couldn't call callback");
-        g_free(jsargs);
-        return;
+        goto out;
     }
 
-    g_callable_info_load_return_type(invoke_info->info, &ret_type);
+    g_callable_info_load_return_type(trampoline->info, &ret_type);
 
-    if (!gjs_value_to_g_argument(invoke_info->context,
+    if (!gjs_value_to_g_argument(trampoline->context,
                                  rval,
                                  &ret_type,
                                  "callback",
@@ -216,21 +198,27 @@ gjs_callback_closure(ffi_cif *cif,
                                  FALSE,
                                  TRUE,
                                  result)) {
-        gjs_throw(invoke_info->context, "Couldn't convert res value");
-        g_free(jsargs);
-        result = NULL;
-        return;
+        goto out;
     }
-    g_free(jsargs);
 
-    gjs_callback_invoke_info_free_content(invoke_info);
-    invoke_info->in_use = FALSE;
+    success = TRUE;
+
+out:
+    if (!success) {
+        gjs_log_exception (trampoline->context, NULL);
+
+        /* Fill in the result with some hopefully neutral value */
+        g_callable_info_load_return_type(trampoline->info, &ret_type);
+        gjs_g_argument_init_default (trampoline->context, &ret_type, result);
+    }
+
+    if (trampoline->scope == GI_SCOPE_TYPE_ASYNC) {
+        completed_trampolines = g_slist_prepend(completed_trampolines, trampoline);
+    }
 }
 
-/* For GI_SCOPE_TYPE_NOTIFIED callback we provide our own
- * ffi closure so we can easily get the user data instead
- * of always having to map the user_data argument on the
- * invokation side.
+/* The global entry point for any invocations of GDestroyNotify;
+ * look up the callback through the user_data and then free it.
  */
 static void
 gjs_destroy_notify_callback_closure(ffi_cif *cif,
@@ -238,223 +226,135 @@ gjs_destroy_notify_callback_closure(ffi_cif *cif,
                                     void **args,
                                     void *data)
 {
-    GjsCallbackInvokeInfo *info = (GjsCallbackInvokeInfo *)data;
+    GjsCallbackTrampoline *trampoline = *(void**)(args[0]);
 
-    info->in_use = FALSE;
-
-    gjs_callback_invoke_info_free_content(info);
+    g_assert(trampoline);
+    gjs_callback_trampoline_free(trampoline);
 }
 
-static gboolean
-gjs_destroy_notify_create(GjsCallbackInvokeInfo *info)
+/* Called when we first see a function that uses a callback */
+static void
+gjs_init_callback_statics ()
 {
-    static const ffi_type *destroy_notify_args[] = {&ffi_type_pointer};
-    ffi_status status;
+    if (G_LIKELY(trampoline_globals_initialized))
+      return;
+    trampoline_globals_initialized = TRUE;
 
-    g_assert(info);
+    global_destroy_trampoline.info = g_irepository_find_by_name(NULL, "GLib", "DestroyNotify");
+    g_assert(global_destroy_trampoline.info != NULL);
+    g_assert(g_base_info_get_type(global_destroy_trampoline.info) == GI_INFO_TYPE_CALLBACK);
 
-    info->need_free_arg_types = FALSE;
 
-    info->closure = mmap(NULL, sizeof (ffi_closure),
-                         PROT_EXEC | PROT_READ | PROT_WRITE,
-                         MAP_ANON | MAP_PRIVATE, -1, sysconf (_SC_PAGE_SIZE));
-    if (!info->closure) {
-        gjs_throw(info->context, "mmap failed\n");
-        return FALSE;
-    }
-
-    info->cif = g_slice_new(ffi_cif);
-
-    status = ffi_prep_cif(info->cif, FFI_DEFAULT_ABI,
-                          1, &ffi_type_void,
-                          (ffi_type**)destroy_notify_args);
-    if (status != FFI_OK) {
-        gjs_throw(info->context, "ffi_prep_cif failed: %d\n", status);
-        munmap(info->closure, sizeof (ffi_closure));
-        return FALSE;
-    }
-
-    status = ffi_prep_closure(info->closure, info->cif,
-                              gjs_destroy_notify_callback_closure, info);
-    if (status != FFI_OK) {
-        gjs_throw(info->context, "ffi_prep_cif failed: %d\n", status);
-        munmap(info->closure, sizeof (ffi_closure));
-        return FALSE;
-    }
-
-    if (mprotect(info->closure, sizeof (ffi_closure), PROT_READ | PROT_EXEC) == -1) {
-        gjs_throw(info->context, "ffi_prep_closure failed: %s\n", strerror(errno));
-        munmap(info->closure, sizeof (ffi_closure));
-        return FALSE;
-    }
-
-    info->in_use = TRUE;
-
-    return TRUE;
+    global_destroy_trampoline.closure = g_callable_info_prepare_closure(global_destroy_trampoline.info,
+                                                                        &global_destroy_trampoline.cif,
+                                                                        gjs_destroy_notify_callback_closure,
+                                                                        NULL);
 }
 
-static GjsCallbackInvokeInfo*
-gjs_callback_invoke_prepare(JSContext      *context,
+static GjsCallbackTrampoline*
+gjs_callback_trampoline_new(JSContext      *context,
                             jsval           function,
-                            GICallableInfo *callable_info)
+                            GICallableInfo *callable_info,
+                            GIScopeType     scope,
+                            void          **destroy_notify)
 {
-    GjsCallbackInvokeInfo *invoke_info;
+    GjsCallbackTrampoline *trampoline;
+
+    if (function == JSVAL_NULL) {
+        *destroy_notify = NULL;
+        return NULL;
+    }
 
     g_assert(JS_TypeOfValue(context, function) == JSTYPE_FUNCTION);
 
-    invoke_info = g_slice_new0(GjsCallbackInvokeInfo);
-    invoke_info->context = context;
-    invoke_info->info = callable_info;
+    trampoline = g_slice_new(GjsCallbackTrampoline);
+    trampoline->context = context;
+    trampoline->info = callable_info;
+    g_base_info_ref((GIBaseInfo*)trampoline->info);
+    trampoline->js_function = function;
+    JS_AddValueRoot(context, &trampoline->js_function);
+    trampoline->closure = g_callable_info_prepare_closure(callable_info, &trampoline->cif,
+                                                          gjs_callback_closure, trampoline);
 
-    g_base_info_ref((GIBaseInfo*)invoke_info->info);
-    invoke_info->need_free_arg_types = TRUE;
-    invoke_info->function = function;
-    invoke_info->cif = g_slice_new0(ffi_cif);
-    invoke_info->in_use = TRUE;
-    invoke_info->callback_info.arguments = NULL;
-    invoke_info->callback_info.invoke_infos = NULL;
-    invoke_info->closure = g_callable_info_prepare_closure(callable_info, invoke_info->cif,
-                                                           gjs_callback_closure, invoke_info);
+    trampoline->scope = scope;
+    if (scope == GI_SCOPE_TYPE_NOTIFIED) {
+        *destroy_notify = global_destroy_trampoline.closure;
+    } else {
+        *destroy_notify = NULL;
+    }
 
-    g_assert(invoke_info->closure);
-
-    return invoke_info;
+    return trampoline;
 }
 
-static void
-gjs_callback_info_free (gpointer data)
+static JSBool
+init_callback_args_for_invocation(JSContext               *context,
+                                  Function                *function,
+                                  guint8                   n_args,
+                                  uintN                    js_argc,
+                                  jsval                   *js_argv,
+                                  GjsCallbackTrampoline  **trampoline_out,
+                                  void                   **destroy_notify_out)
 {
-    g_slice_free(GjsCallbackInfo, data);
-}
+    GIArgInfo callback_arg;
+    GITypeInfo callback_type;
+    GITypeInfo *callback_info;
+    GIScopeType scope;
+    gboolean found_js_function;
+    jsval js_function;
+    guint8 i, js_argv_pos;
 
-static inline void
-gjs_callback_info_add_argument(JSContext       *context,
-                               GjsCallbackInfo *callback_info,
-                               jsval            arg)
-{
-    jsval *v;
-    /*JS_AddRoot/JS_RemoveRoot pair require same pointer*/
-    v = g_slice_new(jsval);
-    *v = arg;
-    JS_AddRoot(context, v);
-    callback_info->arguments = g_slist_prepend(callback_info->arguments, v);
-}
+    if (function->callback_index == GJS_ARG_INDEX_INVALID) {
+        *trampoline_out = *destroy_notify_out = NULL;
+        return JS_TRUE;
+    }
 
-static inline gboolean
-gjs_callback_from_arguments(JSContext *context,
-                            GIBaseInfo* interface_info,
-                            GIArgInfo *arg_info,
-                            guint8 current_arg_pos,
-                            guint8 n_args,
-                            guint8 *argv_pos,
-                            uintN argc,
-                            jsval *argv,
-                            GSList **all_invoke_infos,
-                            GSList **data_for_notify,
-                            GSList **call_free_list,
-                            gpointer *closure)
-{
-    GjsCallbackInvokeInfo *invoke_info;
-    GSList *l;
-    gboolean is_notify = FALSE;
-    GjsCallbackInfo *callback_info = NULL;
+    g_callable_info_load_arg( (GICallableInfo*) function->info, function->callback_index, &callback_arg);
+    scope = g_arg_info_get_scope(&callback_arg);
+    g_arg_info_load_type(&callback_arg, &callback_type);
+    g_assert(g_type_info_get_tag(&callback_type) == GI_TYPE_TAG_INTERFACE);
+    callback_info = g_type_info_get_interface(&callback_type);
+    g_assert(g_base_info_get_type(callback_info) == GI_INFO_TYPE_CALLBACK);
 
-    for (l = *data_for_notify; l; l = l->next) {
-        GjsCallbackInfo *callback_info = l->data;
-        if (callback_info->arg_index != current_arg_pos)
+    /* Find the JS function passed for the callback */
+    found_js_function = FALSE;
+    js_function = JSVAL_NULL;
+    for (i = 0, js_argv_pos = 0; i < n_args; i++) {
+        if (i == function->callback_index) {
+            js_function = js_argv[js_argv_pos];
+            found_js_function = TRUE;
+            break;
+        } else if (i == function->user_data_index
+                   || i == function->destroy_notify_index) {
             continue;
-
-        invoke_info = g_slice_new0(GjsCallbackInvokeInfo);
-        invoke_info->context = context;
-
-        if (!gjs_destroy_notify_create(invoke_info)) {
-            g_slice_free(GjsCallbackInvokeInfo, invoke_info);
-            return FALSE;
         }
-
-        if (*argv_pos < argc) {
-            gjs_callback_info_add_argument(context, callback_info, argv[*argv_pos]);
-            (*argv_pos)--;
-        }
-        is_notify = TRUE;
-        invoke_info->callback_info = *callback_info;
-        *all_invoke_infos = g_slist_prepend(*all_invoke_infos, invoke_info);
-        break;
+        js_argv_pos++;
     }
 
-    if (is_notify)
-        goto out;
-
-    g_assert_cmpuint(*argv_pos, <, argc);
-    if (JSVAL_IS_NULL(argv[*argv_pos]) || JSVAL_IS_VOID(argv[*argv_pos])) {
-        *closure = NULL;
-        return TRUE;
+    if (!found_js_function
+        || !(js_function == JSVAL_NULL || JS_TypeOfValue(context, js_function) == JSTYPE_FUNCTION)) {
+        gjs_throw(context, "Error invoking %s.%s: Invalid callback given for argument %s",
+                  g_base_info_get_namespace( (GIBaseInfo*) function->info),
+                  g_base_info_get_name( (GIBaseInfo*) function->info),
+                  g_base_info_get_name( (GIBaseInfo*) &callback_arg));
+        g_base_info_unref( (GIBaseInfo*) callback_info);
+        return JS_FALSE;
     }
 
-    invoke_info = gjs_callback_invoke_prepare(context,
-                                              argv[*argv_pos],
-                                              (GICallableInfo*)interface_info);
-
-    switch (g_arg_info_get_scope(arg_info)) {
-        case GI_SCOPE_TYPE_CALL:
-            *call_free_list = g_slist_prepend(*call_free_list, invoke_info);
-            break;
-        case GI_SCOPE_TYPE_NOTIFIED:
-            g_assert(callback_info == NULL);
-            callback_info = g_slice_new0(GjsCallbackInfo);
-
-            g_assert(g_arg_info_get_destroy(arg_info) < n_args);
-
-            gjs_callback_info_add_argument(context, callback_info, argv[*argv_pos]);
-            callback_info->arg_index = g_arg_info_get_destroy(arg_info);
-
-            callback_info->invoke_infos = g_slist_prepend(callback_info->invoke_infos, invoke_info);
-            *data_for_notify = g_slist_prepend(*data_for_notify, callback_info);
-            break;
-        case GI_SCOPE_TYPE_ASYNC:
-            gjs_callback_info_add_argument(context, &invoke_info->callback_info, argv[*argv_pos]);
-
-            *all_invoke_infos = g_slist_prepend(*all_invoke_infos, invoke_info);
-            break;
-        default:
-            gjs_throw(context, "Unknown callback scope");
-            gjs_callback_invoke_info_free(invoke_info);
-            return FALSE;
-    }
-
-out:
-    *closure = invoke_info->closure;
-
-    return TRUE;
-}
-
-static void
-gjs_free_unused_invoke_infos(GSList **invoke_infos)
-{
-    GSList *l, *node_for_delete = NULL;
-
-    for (l = *invoke_infos; l; l = l->next) {
-        GjsCallbackInvokeInfo *info = l->data;
-
-         if (info->in_use)
-             continue;
-
-         gjs_callback_invoke_info_free(info);
-         node_for_delete = g_slist_prepend(node_for_delete, l);
-    }
-    for (l = node_for_delete; l; l = l->next) {
-        *invoke_infos = g_slist_delete_link(*invoke_infos, l->data);
-     }
-    g_slist_free(node_for_delete);
+    *trampoline_out = gjs_callback_trampoline_new(context, js_function,
+                                                  (GICallbackInfo*) callback_info,
+                                                  scope,
+                                                  destroy_notify_out);
+    g_base_info_unref( (GIBaseInfo*) callback_info);
+    return JS_TRUE;
 }
 
 static JSBool
 gjs_invoke_c_function(JSContext      *context,
                       Function       *function,
                       JSObject       *obj, /* "this" object */
-                      uintN           argc,
-                      jsval          *argv,
-                      jsval          *rval)
+                      uintN           js_argc,
+                      jsval          *js_argv,
+                      jsval          *js_rval)
 {
     /* These first four are arrays which hold argument pointers.
      * @in_arg_cvalues: C values which are passed on input (in or inout)
@@ -474,7 +374,7 @@ gjs_invoke_c_function(JSContext      *context,
     GArgument return_value;
 
     guint8 processed_in_args;
-    guint8 n_args, i, argv_pos;
+    guint8 n_args, i, js_argv_pos;
     guint8 in_args_pos, out_args_pos, inout_args_pos;
     guint8 in_args_len, out_args_len, inout_args_len;
     guint8 can_throw_gerror, did_throw_gerror;
@@ -485,38 +385,48 @@ gjs_invoke_c_function(JSContext      *context,
     gboolean is_method;
     GITypeInfo return_info;
     GITypeTag return_tag;
-    jsval *return_values;
-    guint8 next_rval;
+    jsval *return_values = NULL;
+    guint8 next_rval = 0; /* index into return_values */
+    GSList *iter;
+    GjsCallbackTrampoline *callback_trampoline;
+    void *destroy_notify;
 
-    GSList *call_free_list = NULL; /* list of GjsCallbackInvokeInfo* */
-    GSList *data_for_notify = NULL; /* list of GjsCallbackInfo* */
-    GSList *callback_arg_indices = NULL; /* list of int */
-    static GSList *invoke_infos = NULL;
-
-    /* For async/notify callbacks we need to clean up some of the closure
-     * data in follow up calls due to the way mmap works. ffi closures works
-     * by rewriting the function trampolines, eg it modifies the page where
-     * the function is mapped, that page cannot be released/mumapped as long
-     * as we're currently using it, eg executing it. And as soon as we leave
-     * that function we loose control, it goes back into the control of the
-     * library where the function is called.
+    /* Because we can't free a closure while we're in it, we defer
+     * freeing until the next time a C function is invoked.  What
+     * we should really do instead is queue it for a GC thread.
      */
-    gjs_free_unused_invoke_infos(&invoke_infos);
+    if (completed_trampolines) {
+        for (iter = completed_trampolines; iter; iter = iter->next) {
+            GjsCallbackTrampoline *trampoline = iter->data;
+            gjs_callback_trampoline_free(trampoline);
+        }
+        g_slist_free(completed_trampolines);
+        completed_trampolines = NULL;
+    }
 
     flags = g_function_info_get_flags(function->info);
     is_method = (flags & GI_FUNCTION_IS_METHOD) != 0;
     can_throw_gerror = (flags & GI_FUNCTION_THROWS) != 0;
+    n_args = g_callable_info_get_n_args( (GICallableInfo*) function->info);
 
     /* We allow too many args; convenient for re-using a function as a callback.
      * But we don't allow too few args, since that would break.
      */
-    if (argc < function->expected_js_argc) {
+    if (js_argc < function->expected_js_argc) {
         gjs_throw(context, "Too few arguments to %s %s.%s expected %d got %d",
                   is_method ? "method" : "function",
                   g_base_info_get_namespace( (GIBaseInfo*) function->info),
                   g_base_info_get_name( (GIBaseInfo*) function->info),
                   function->expected_js_argc,
-                  argc);
+                  js_argc);
+        return JS_FALSE;
+    }
+
+    /* Check if we have a callback; if so, process all the arguments (callback, destroy_notify, user_data)
+     * at once to avoid having special cases in the main loop below.
+     */
+    if (!init_callback_args_for_invocation(context, function, n_args, js_argc, js_argv,
+                                           &callback_trampoline, &destroy_notify)) {
         return JS_FALSE;
     }
 
@@ -540,9 +450,7 @@ gjs_invoke_c_function(JSContext      *context,
     in_args_pos = 0; /* index into in_arg_cvalues */
     out_args_pos = 0; /* index into out_arg_cvalues */
     inout_args_pos = 0; /* index into inout_original_arg_cvalues */
-    argv_pos = 0; /* index into argv */
-
-    n_args = g_callable_info_get_n_args( (GICallableInfo*) function->info);
+    js_argv_pos = 0; /* index into argv */
 
     if (is_method) {
         GIBaseInfo *container = g_base_info_get_container((GIBaseInfo *) function->info);
@@ -567,7 +475,7 @@ gjs_invoke_c_function(JSContext      *context,
         GIArgInfo arg_info;
         gboolean arg_removed = FALSE;
 
-        /* gjs_debug(GJS_DEBUG_GFUNCTION, "i: %d in_args_pos: %d argv_pos: %d", i, in_args_pos, argv_pos); */
+        /* gjs_debug(GJS_DEBUG_GFUNCTION, "i: %d in_args_pos: %d argv_pos: %d", i, in_args_pos, js_argv_pos); */
 
         g_callable_info_load_arg( (GICallableInfo*) function->info, i, &arg_info);
         direction = g_arg_info_get_direction(&arg_info);
@@ -586,7 +494,6 @@ gjs_invoke_c_function(JSContext      *context,
             GArgument *in_value;
             GITypeTag type_tag;
             GITypeInfo ainfo;
-            gboolean convert_argument = TRUE;
 
             g_arg_info_load_type(&arg_info, &ainfo);
             type_tag = g_type_info_get_tag(&ainfo);
@@ -594,52 +501,22 @@ gjs_invoke_c_function(JSContext      *context,
             g_assert_cmpuint(in_args_pos, <, in_args_len);
             in_value = &in_arg_cvalues[in_args_pos];
 
-            if (g_slist_find(callback_arg_indices, GUINT_TO_POINTER((guint)i)) != NULL) {
-                g_assert(type_tag == GI_TYPE_TAG_VOID);
+            /* First check for callback stuff */
+            if (i == function->callback_index) {
+                if (callback_trampoline)
+                    in_value->v_pointer = callback_trampoline->closure;
+                else
+                    in_value->v_pointer = NULL;
+            } else if (i == function->user_data_index) {
+                in_value->v_pointer = callback_trampoline;
                 arg_removed = TRUE;
-                convert_argument = FALSE;
-                in_value->v_pointer = NULL;
-            } else if (type_tag == GI_TYPE_TAG_VOID) {
-                /* FIXME: notify/throw saying the callback annotation is wrong */
-                convert_argument = FALSE;
-
-                g_assert_cmpuint(argv_pos, <, argc);
-                in_value->v_pointer = (gpointer)argv[argv_pos];
-            } else if (type_tag == GI_TYPE_TAG_INTERFACE) {
-                GIBaseInfo* interface_info;
-                GIInfoType interface_type;
-
-                interface_info = g_type_info_get_interface(&ainfo);
-
-                g_assert(interface_info != NULL);
-
-                interface_type = g_base_info_get_type(interface_info);
-                if (interface_type == GI_INFO_TYPE_CALLBACK) {
-                    gint user_data_pos;
-
-                    if (!gjs_callback_from_arguments(context, interface_info, &arg_info,
-                                                     i, n_args, &argv_pos, argc, argv,
-                                                     &invoke_infos,
-                                                     &data_for_notify, &call_free_list,
-                                                     &(in_value->v_pointer))) {
-                        failed = TRUE;
-                        break;
-                    }
-                    user_data_pos = g_arg_info_get_closure(&arg_info);
-                    if (is_method)
-                        --user_data_pos;
-                    callback_arg_indices = g_slist_prepend(callback_arg_indices,
-                                                           GINT_TO_POINTER(user_data_pos));
-
-                    convert_argument = FALSE;
-                }
-
-                g_base_info_unref(interface_info);
-            }
-
-            if (convert_argument) {
-                g_assert_cmpuint(argv_pos, <, argc);
-                if (!gjs_value_to_arg(context, argv[argv_pos], &arg_info,
+            } else if (i == function->destroy_notify_index) {
+                in_value->v_pointer = destroy_notify;
+                arg_removed = TRUE;
+            } else {
+                /* Ok, now just convert argument normally */
+                g_assert_cmpuint(js_argv_pos, <, js_argc);
+                if (!gjs_value_to_arg(context, js_argv[js_argv_pos], &arg_info,
                                       in_value)) {
                     failed = TRUE;
                     break;
@@ -658,7 +535,7 @@ gjs_invoke_c_function(JSContext      *context,
             }
 
             if (!arg_removed)
-                ++argv_pos;
+                ++js_argv_pos;
         }
 
         ++in_args_pos;
@@ -699,17 +576,8 @@ gjs_invoke_c_function(JSContext      *context,
         did_throw_gerror = FALSE;
     }
 
-    g_slist_foreach(call_free_list, (GFunc)gjs_callback_invoke_info_free, NULL);
-    g_slist_free(call_free_list);
-    g_slist_foreach(data_for_notify, (GFunc)gjs_callback_info_free, NULL);
-    g_slist_free(data_for_notify);
-    g_slist_free(callback_arg_indices);
+    *js_rval = JSVAL_VOID;
 
-    *rval = JSVAL_VOID;
-
-    next_rval = 0; /* index into return_values */
-
-    return_values = NULL; /* Quiet gcc warning about initialization */
     /* Only process return values if the function didn't throw */
     if (function->js_out_argc > 0 && !did_throw_gerror) {
         return_values = g_newa(jsval, function->js_out_argc);
@@ -738,6 +606,11 @@ gjs_invoke_c_function(JSContext      *context,
     }
 
 release:
+    if (callback_trampoline &&
+        callback_trampoline->scope == GI_SCOPE_TYPE_CALL) {
+        gjs_callback_trampoline_free(callback_trampoline);
+    }
+
     /* We walk over all args, release in args (if allocated) and convert
      * all out args to JS
      */
@@ -838,7 +711,7 @@ release:
          * [return value, out arg 1, out arg 2, ...]
          */
         if (function->js_out_argc == 1) {
-            *rval = return_values[0];
+            *js_rval = return_values[0];
         } else {
             JSObject *array;
             array = JS_NewArrayObject(context,
@@ -847,7 +720,7 @@ release:
             if (array == NULL) {
                 failed = TRUE;
             } else {
-                *rval = OBJECT_TO_JSVAL(array);
+                *js_rval = OBJECT_TO_JSVAL(array);
             }
         }
 
@@ -876,14 +749,14 @@ release:
 static JSBool
 function_call(JSContext *context,
               JSObject  *obj, /* "this" object, not the function object */
-              uintN      argc,
-              jsval     *argv,
+              uintN      js_argc,
+              jsval     *js_argv,
               jsval     *rval)
 {
     Function *priv;
     JSObject *callee;
 
-    callee = JSVAL_TO_OBJECT(JS_ARGV_CALLEE(argv)); /* Callee is the Function object being called */
+    callee = JSVAL_TO_OBJECT(JS_ARGV_CALLEE(js_argv)); /* Callee is the Function object being called */
 
     priv = priv_from_js(context, callee);
     gjs_debug_marshal(GJS_DEBUG_GFUNCTION, "Call callee %p priv %p this obj %p %s", callee, priv,
@@ -893,7 +766,7 @@ function_call(JSContext *context,
     if (priv == NULL)
         return JS_TRUE; /* we are the prototype, or have the wrong class */
 
-    return gjs_invoke_c_function(context, priv, obj, argc, argv, rval);
+    return gjs_invoke_c_function(context, priv, obj, js_argc, js_argv, rval);
 }
 
 /* If we set JSCLASS_CONSTRUCT_PROTOTYPE flag, then this is called on
@@ -926,6 +799,17 @@ function_constructor(JSContext *context,
     return JS_TRUE;
 }
 
+/* Does not actually free storage for structure, just
+ * reverses init_cached_function_data
+ */
+static void
+uninit_cached_function_data (Function *function)
+{
+    if (function->info)
+        g_base_info_unref( (GIBaseInfo*) function->info);
+    g_function_invoker_destroy(&function->invoker);
+}
+
 static void
 function_finalize(JSContext *context,
                   JSObject  *obj)
@@ -938,8 +822,7 @@ function_finalize(JSContext *context,
     if (priv == NULL)
         return; /* we are the prototype, not a real instance, so constructor never called */
 
-    if (priv->info)
-        g_base_info_unref( (GIBaseInfo*) priv->info);
+    uninit_cached_function_data(priv);
 
     GJS_DEC_COUNTER(function);
     g_slice_free(Function, priv);
@@ -1004,28 +887,62 @@ init_cached_function_data (JSContext      *context,
 
     n_args = g_callable_info_get_n_args((GICallableInfo*) info);
 
+    function->callback_index = GJS_ARG_INDEX_INVALID;
+    function->destroy_notify_index = GJS_ARG_INDEX_INVALID;
+    function->user_data_index = GJS_ARG_INDEX_INVALID;
+
     for (i = 0; i < n_args; i++) {
         GIDirection direction;
         GIArgInfo arg_info;
+        GITypeInfo type_info;
         guint8 destroy, closure;
+        GITypeTag type_tag;
 
         g_callable_info_load_arg((GICallableInfo*) info, i, &arg_info);
+        g_arg_info_load_type(&arg_info, &type_info);
+        type_tag = g_type_info_get_tag(&type_info);
+        if (type_tag == GI_TYPE_TAG_INTERFACE) {
+            GIBaseInfo* interface_info;
+            GIInfoType interface_type;
+
+            interface_info = g_type_info_get_interface(&type_info);
+            interface_type = g_base_info_get_type(interface_info);
+            if (interface_type == GI_INFO_TYPE_CALLBACK &&
+                i != function->destroy_notify_index) {
+                if (function->callback_index != GJS_ARG_INDEX_INVALID) {
+                    gjs_throw(context, "Function %s.%s has multiple callbacks, not supported",
+                              g_base_info_get_namespace( (GIBaseInfo*) info),
+                              g_base_info_get_name( (GIBaseInfo*) info));
+                    g_base_info_unref(interface_info);
+                    return FALSE;
+                }
+                function->callback_index = i;
+                gjs_init_callback_statics();
+            }
+            g_base_info_unref(interface_info);
+        }
         destroy = g_arg_info_get_destroy(&arg_info);
-        if (is_method)
-            --destroy;
         closure = g_arg_info_get_closure(&arg_info);
-        if (is_method)
-            --closure;
         direction = g_arg_info_get_direction(&arg_info);
 
         if (destroy > 0 && destroy < n_args) {
             function->expected_js_argc -= 1;
-            function->n_destroy_notifies += 1;
+            if (function->destroy_notify_index != GJS_ARG_INDEX_INVALID) {
+                gjs_throw(context, "Function %s has multiple GDestroyNotify, not supported",
+                          g_base_info_get_name((GIBaseInfo*)info));
+                return FALSE;
+            }
+            function->destroy_notify_index = destroy;
         }
 
         if (closure > 0 && closure < n_args) {
             function->expected_js_argc -= 1;
-            function->n_user_data += 1;
+            if (function->user_data_index != GJS_ARG_INDEX_INVALID) {
+                gjs_throw(context, "Function %s has multiple user_data arguments, not supported",
+                          g_base_info_get_name((GIBaseInfo*)info));
+                return FALSE;
+            }
+            function->user_data_index = closure;
         }
 
         if (direction == GI_DIRECTION_IN || direction == GI_DIRECTION_INOUT)
@@ -1036,6 +953,16 @@ init_cached_function_data (JSContext      *context,
             function->inout_argc += 1;
     }
 
+
+    if (function->callback_index != GJS_ARG_INDEX_INVALID
+        && function->destroy_notify_index != GJS_ARG_INDEX_INVALID
+        && function->user_data_index == GJS_ARG_INDEX_INVALID) {
+        gjs_throw(context, "Function %s.%s has a GDestroyNotify but no user_data, not supported",
+                  g_base_info_get_namespace( (GIBaseInfo*) info),
+                  g_base_info_get_name( (GIBaseInfo*) info));
+        return JS_FALSE;
+    }
+
     function->info = info;
 
     g_base_info_ref((GIBaseInfo*) function->info);
@@ -1113,10 +1040,13 @@ gjs_define_function(JSContext      *context,
     JSContext *load_context;
 
     load_context = gjs_runtime_get_load_context(JS_GetRuntime(context));
+    JS_BeginRequest(load_context);
 
     function = function_new(load_context, info);
     if (function == NULL) {
         gjs_move_exception(load_context, context);
+
+        JS_EndRequest(load_context);
         return NULL;
     }
 
@@ -1126,9 +1056,12 @@ gjs_define_function(JSContext      *context,
                            NULL, NULL,
                            GJS_MODULE_PROP_FLAGS)) {
         gjs_debug(GJS_DEBUG_GFUNCTION, "Failed to define function");
+
+        JS_EndRequest(load_context);
         return NULL;
     }
 
+    JS_EndRequest(load_context);
     return function;
 }
 
@@ -1142,10 +1075,13 @@ gjs_invoke_c_function_uncached (JSContext      *context,
                                 jsval          *rval)
 {
   Function function;
+  JSBool result;
 
   memset (&function, 0, sizeof (Function));
   if (!init_cached_function_data (context, &function, info))
     return JS_FALSE;
 
-  return gjs_invoke_c_function (context, &function, obj, argc, argv, rval);
+  result = gjs_invoke_c_function (context, &function, obj, argc, argv, rval);
+  uninit_cached_function_data (&function);
+  return result;
 }
