@@ -31,8 +31,10 @@
 #include "function.h"
 #include "value.h"
 #include "keep-alive.h"
+#include "gjs_gi_trace.h"
 
 #include <gjs/gjs.h>
+#include <gjs/compat.h>
 
 #include <util/log.h>
 
@@ -119,7 +121,7 @@ init_g_param_from_property(JSContext  *context,
 static JSBool
 object_instance_get_prop(JSContext *context,
                          JSObject  *obj,
-                         jsval      id,
+                         jsid       id,
                          jsval     *value_p)
 {
     ObjectInstance *priv;
@@ -128,7 +130,7 @@ object_instance_get_prop(JSContext *context,
     GParamSpec *param;
     GValue gvalue = { 0, };
 
-    if (!gjs_get_string_id(id, &name))
+    if (!gjs_get_string_id(context, id, &name))
         return JS_TRUE; /* not resolved, but no error */
 
     priv = priv_from_js(context, obj);
@@ -176,14 +178,14 @@ object_instance_get_prop(JSContext *context,
 static JSBool
 object_instance_set_prop(JSContext *context,
                          JSObject  *obj,
-                         jsval      id,
+                         jsid       id,
                          jsval     *value_p)
 {
     ObjectInstance *priv;
     const char *name;
     GParameter param = { NULL, { 0, }};
 
-    if (!gjs_get_string_id(id, &name))
+    if (!gjs_get_string_id(context, id, &name))
         return JS_TRUE; /* not resolved, but no error */
 
     priv = priv_from_js(context, obj);
@@ -237,7 +239,7 @@ object_instance_set_prop(JSContext *context,
 static JSBool
 object_instance_new_resolve(JSContext *context,
                             JSObject  *obj,
-                            jsval      id,
+                            jsid       id,
                             uintN      flags,
                             JSObject **objp)
 {
@@ -246,7 +248,7 @@ object_instance_new_resolve(JSContext *context,
 
     *objp = NULL;
 
-    if (!gjs_get_string_id(id, &name))
+    if (!gjs_get_string_id(context, id, &name))
         return JS_TRUE; /* not resolved, but no error */
 
     priv = priv_from_js(context, obj);
@@ -457,27 +459,23 @@ object_instance_props_to_g_parameters(JSContext   *context,
         return JS_FALSE;
     }
 
-    prop_id = JSVAL_VOID;
+    prop_id = JSID_VOID;
     if (!JS_NextProperty(context, iter, &prop_id))
         return JS_FALSE;
 
-    if (prop_id != JSVAL_VOID) {
+    if (!JSID_IS_VOID(prop_id)) {
         gparams = g_array_new(/* nul term */ FALSE, /* clear */ TRUE,
                               sizeof(GParameter));
     } else {
         return JS_TRUE;
     }
 
-    while (prop_id != JSVAL_VOID) {
-        jsval nameval;
+    while (!JSID_IS_VOID(prop_id)) {
         const char *name;
         jsval value;
         GParameter gparam = { NULL, { 0, }};
 
-        if (!JS_IdToValue(context, prop_id, &nameval))
-            goto free_array_and_fail;
-
-        if (!gjs_get_string_id(nameval, &name))
+        if (!gjs_get_string_id(context, prop_id, &name))
             goto free_array_and_fail;
 
         if (!gjs_object_require_property(context, props, "property list", name, &value))
@@ -499,7 +497,7 @@ object_instance_props_to_g_parameters(JSContext   *context,
 
         g_array_append_val(gparams, gparam);
 
-        prop_id = JSVAL_VOID;
+        prop_id = JSID_VOID;
         if (!JS_NextProperty(context, iter, &prop_id))
             goto free_array_and_fail;
     }
@@ -765,6 +763,9 @@ object_instance_constructor(JSContext *context,
         gjs_debug_lifecycle(GJS_DEBUG_GOBJECT,
                             "JSObject created with GObject %p %s",
                             priv->gobj, g_type_name_from_instance((GTypeInstance*) priv->gobj));
+
+        TRACE(GJS_OBJECT_PROXY_NEW(priv, priv->gobj, g_base_info_get_namespace ( (GIBaseInfo*) priv->info),
+                                    g_base_info_get_name ( (GIBaseInfo*) priv->info) ));
     }
 
     return JS_TRUE;
@@ -786,6 +787,9 @@ object_instance_finalize(JSContext *context,
     if (priv == NULL)
         return; /* we are the prototype, not a real instance, so constructor never called */
 
+    TRACE(GJS_OBJECT_PROXY_FINALIZE(priv, priv->gobj, g_base_info_get_namespace ( (GIBaseInfo*) priv->info),
+                                    g_base_info_get_name ( (GIBaseInfo*) priv->info) ));
+
     if (priv->gobj) {
         g_assert(priv->gobj->ref_count > 0);
         set_js_obj(context, priv->gobj, NULL);
@@ -851,7 +855,7 @@ real_connect_func(JSContext *context,
     const char *signal_name;
     GQuark signal_detail;
 
-    *retval = INT_TO_JSVAL(0);
+    *retval = JSID_VOID;
 
     priv = priv_from_js(context, obj);
     gjs_debug_gsignal("connect obj %p priv %p argc %d", obj, priv, argc);
@@ -1371,9 +1375,9 @@ gjs_define_object_class(JSContext     *context,
                                        /* number of constructor args */
                                        0,
                                        /* props of prototype */
-                                       &gjs_object_instance_proto_props[0],
+                                       parent_proto ? NULL : &gjs_object_instance_proto_props[0],
                                        /* funcs of prototype */
-                                       &gjs_object_instance_proto_funcs[0],
+                                       parent_proto ? NULL : &gjs_object_instance_proto_funcs[0],
                                        /* props of constructor, MyConstructor.myprop */
                                        NULL,
                                        /* funcs of constructor, MyConstructor.myfunc() */
