@@ -31,8 +31,10 @@
 #include "union.h"
 #include "enumeration.h"
 #include "arg.h"
+#include "foreign.h"
 
 #include <gjs/mem.h>
+#include <gjs/compat.h>
 
 #include <util/log.h>
 #include <util/misc.h>
@@ -63,11 +65,16 @@ resolve_namespace_object(JSContext  *context,
     JSObject *versions;
     jsval version_val;
     const char *version;
+    JSObject *result;
 
     load_context = gjs_runtime_get_load_context(JS_GetRuntime(context));
+    JS_BeginRequest(load_context);
+
     if (!gjs_object_require_property(load_context, repo_obj, "GI repository object", "versions", &versions_val) ||
         !JSVAL_IS_OBJECT(versions_val)) {
         gjs_throw(context, "No 'versions' property in GI repository object");
+
+        JS_EndRequest(load_context);
         return NULL;
     }
 
@@ -88,6 +95,7 @@ resolve_namespace_object(JSContext  *context,
                   "Requiring %s, version %s: %s",
                   ns_name, version?version:"none", error->message);
         g_error_free(error);
+        JS_EndRequest(load_context);
         return JS_FALSE;
     }
 
@@ -95,7 +103,9 @@ resolve_namespace_object(JSContext  *context,
      * with the given namespace name, pointing to that namespace
      * in the repo.
      */
-    return gjs_define_ns(context, repo_obj, ns_name, repo);
+    result = gjs_define_ns(context, repo_obj, ns_name, repo);
+    JS_EndRequest(load_context);
+    return result;
 }
 
 /*
@@ -114,7 +124,7 @@ resolve_namespace_object(JSContext  *context,
 static JSBool
 repo_new_resolve(JSContext *context,
                  JSObject  *obj,
-                 jsval      id,
+                 jsid       id,
                  uintN      flags,
                  JSObject **objp)
 {
@@ -124,7 +134,7 @@ repo_new_resolve(JSContext *context,
 
     *objp = NULL;
 
-    if (!gjs_get_string_id(id, &name))
+    if (!gjs_get_string_id(context, id, &name))
         return JS_TRUE; /* not resolved, but no error */
 
     /* let Object.prototype resolve these */
@@ -139,11 +149,14 @@ repo_new_resolve(JSContext *context,
         return JS_TRUE; /* we are the prototype, or have the wrong class */
 
     load_context = gjs_runtime_get_load_context(JS_GetRuntime(context));
+    JS_BeginRequest(load_context);
     resolve_namespace_object(load_context, obj, name);
     if (gjs_move_exception(load_context, context)) {
+        JS_EndRequest(load_context);
         return JS_FALSE;
     } else {
         *objp = obj; /* store the object we defined the prop in */
+        JS_EndRequest(load_context);
         return JS_TRUE;
     }
 }
@@ -421,7 +434,6 @@ gjs_define_info(JSContext  *context,
 #if GJS_VERBOSE_ENABLE_GI_USAGE
     _gjs_log_info_usage(info);
 #endif
-
     switch (g_base_info_get_type(info)) {
     case GI_INFO_TYPE_FUNCTION:
         {
@@ -432,8 +444,15 @@ gjs_define_info(JSContext  *context,
         }
         break;
     case GI_INFO_TYPE_OBJECT:
-        if (!gjs_define_object_class(context, in_object, G_TYPE_INVALID, (GIObjectInfo*) info, NULL, NULL))
-            return JS_FALSE;
+        {
+            GType gtype;
+            GIBaseInfo *info_for_gtype;
+            gtype = g_registered_type_info_get_g_type((GIRegisteredTypeInfo*)info);
+            if (!gjs_define_object_class(context, in_object, gtype, NULL, NULL, &info_for_gtype))
+                return JS_FALSE;
+            g_assert(g_base_info_equal(info, info_for_gtype));
+            g_base_info_unref(info_for_gtype);
+        }
         break;
     case GI_INFO_TYPE_STRUCT:
     case GI_INFO_TYPE_BOXED:
@@ -500,6 +519,7 @@ gjs_lookup_namespace_object_by_name(JSContext      *context,
      */
 
     load_context = gjs_runtime_get_load_context(JS_GetRuntime(context));
+    JS_BeginRequest(load_context);
     global = JS_GetGlobalObject(load_context);
 
     importer = JSVAL_VOID;
@@ -507,7 +527,7 @@ gjs_lookup_namespace_object_by_name(JSContext      *context,
         !JSVAL_IS_OBJECT(importer)) {
         gjs_log_exception(load_context, NULL);
         gjs_throw(context, "No imports property in global object");
-        return NULL;
+        goto fail;
     }
 
     girepository = JSVAL_VOID;
@@ -516,20 +536,26 @@ gjs_lookup_namespace_object_by_name(JSContext      *context,
         !JSVAL_IS_OBJECT(girepository)) {
         gjs_log_exception(load_context, NULL);
         gjs_throw(context, "No gi property in importer");
-        return NULL;
+        goto fail;
     }
 
     repo_obj = JSVAL_TO_OBJECT(girepository);
 
-    if (!gjs_object_require_property(context, repo_obj, "GI repository object", ns, &ns_obj))
-        return NULL;
+    if (!gjs_object_require_property(context, repo_obj, "GI repository object", ns, &ns_obj)) {
+        goto fail;
+    }
 
     if (!JSVAL_IS_OBJECT(ns_obj)) {
         gjs_throw(context, "Namespace '%s' is not an object?", ns);
-        return NULL;
+        goto fail;
     }
 
+    JS_EndRequest(load_context);
     return JSVAL_TO_OBJECT(ns_obj);
+
+ fail:
+    JS_EndRequest(load_context);
+    return NULL;
 }
 
 const char*
