@@ -1,6 +1,6 @@
 /* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
 /*
- * Copyright (c) 2008  LiTL, LLC
+ * Copyright (c) 2008  litl, LLC
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -31,7 +31,20 @@
 #include "param.h"
 #include "object.h"
 #include "boxed.h"
-#include <gjs/jsapi-util.h>
+#include "union.h"
+#include <gjs/gjs.h>
+#include <gjs/compat.h>
+
+#include <girepository.h>
+
+static GType _array_type = G_TYPE_INVALID;
+static GType _byte_array_type = G_TYPE_INVALID;
+static GType _ptr_array_type = G_TYPE_INVALID;
+
+static JSBool gjs_value_from_g_value_internal(JSContext    *context,
+                                              jsval        *value_p,
+                                              const GValue *gvalue,
+                                              gboolean      no_copy);
 
 static void
 closure_marshal(GClosure        *closure,
@@ -46,6 +59,7 @@ closure_marshal(GClosure        *closure,
     jsval *argv;
     jsval rval;
     int i;
+    GSignalQuery signal_query = { 0, };
 
     gjs_debug_marshal(GJS_DEBUG_GCLOSURE,
                       "Marshal closure %p",
@@ -57,17 +71,48 @@ closure_marshal(GClosure        *closure,
         return;
     }
 
+    JS_BeginRequest(context);
+
     argc = n_param_values;
     argv = g_newa(jsval, n_param_values);
     rval = JSVAL_VOID;
 
     gjs_set_values(context, argv, argc, JSVAL_VOID);
     gjs_root_value_locations(context, argv, argc);
-    JS_AddRoot(context, &rval);
+    JS_AddValueRoot(context, &rval);
+
+    if (marshal_data) {
+        /* we are used for a signal handler */
+        guint signal_id;
+
+        signal_id = GPOINTER_TO_UINT(marshal_data);
+
+        g_signal_query(signal_id, &signal_query);
+
+        if (!signal_query.signal_id) {
+            gjs_debug(GJS_DEBUG_GCLOSURE,
+                      "Signal handler being called on invalid signal");
+            goto cleanup;
+        }
+
+        if (signal_query.n_params + 1 != n_param_values) {
+            gjs_debug(GJS_DEBUG_GCLOSURE,
+                      "Signal handler being called with wrong number of parameters");
+            goto cleanup;
+        }
+    }
 
     for (i = 0; i < argc; ++i) {
         const GValue *gval = &param_values[i];
-        if (!gjs_value_from_g_value(context, &argv[i], gval)) {
+        gboolean no_copy;
+
+        no_copy = FALSE;
+
+        if (i >= 1 && signal_query.signal_id) {
+            no_copy = (signal_query.param_types[i - 1] & G_SIGNAL_TYPE_STATIC_SCOPE) != 0;
+        }
+
+        if (!gjs_value_from_g_value_internal(context, &argv[i], gval, no_copy)) {
             gjs_debug(GJS_DEBUG_GCLOSURE,
                       "Unable to convert arg %d in order to invoke closure",
                       i);
@@ -94,7 +139,23 @@ closure_marshal(GClosure        *closure,
 
  cleanup:
     gjs_unroot_value_locations(context, argv, argc);
-    JS_RemoveRoot(context, &rval);
+    JS_RemoveValueRoot(context, &rval);
+    JS_EndRequest(context);
+}
+
+GClosure*
+gjs_closure_new_for_signal(JSContext  *context,
+                           JSObject   *callable,
+                           const char *description,
+                           guint       signal_id)
+{
+    GClosure *closure;
+
+    closure = gjs_closure_new(context, callable, description);
+
+    g_closure_set_meta_marshal(closure, GUINT_TO_POINTER(signal_id), closure_marshal);
+
+    return closure;
 }
 
 GClosure*
@@ -136,10 +197,11 @@ gjs_value_guess_g_type(jsval value)
     return G_TYPE_INVALID;
 }
 
-JSBool
-gjs_value_to_g_value(JSContext    *context,
-                     jsval         value,
-                     GValue       *gvalue)
+static JSBool
+gjs_value_to_g_value_internal(JSContext    *context,
+                              jsval         value,
+                              GValue       *gvalue,
+                              gboolean      no_copy)
 {
     GType gtype;
 
@@ -279,6 +341,42 @@ gjs_value_to_g_value(JSContext    *context,
         }
 
         g_value_set_object(gvalue, gobj);
+    } else if (gtype == G_TYPE_STRV) {
+        if (JSVAL_IS_NULL(value)) {
+            /* do nothing */
+        } else if (gjs_object_has_property(context,
+                                           JSVAL_TO_OBJECT(value),
+                                           "length")) {
+            jsval length_value;
+            guint32 length;
+
+            if (!gjs_object_require_property(context,
+                                             JSVAL_TO_OBJECT(value), NULL,
+                                             "length",
+                                             &length_value) ||
+                !JS_ValueToECMAUint32(context, length_value, &length)) {
+                gjs_throw(context,
+                          "Wrong type %s; strv expected",
+                          gjs_get_type_name(value));
+                return JS_FALSE;
+            } else {
+                void *result;
+                char **strv;
+
+                if (!gjs_array_to_strv (context,
+                                        value,
+                                        length, &result))
+                    return JS_FALSE;
+                /* cast to strv in a separate step to avoid type-punning */
+                strv = result;
+                g_value_take_boxed (gvalue, strv);
+            }
+        } else {
+            gjs_throw(context,
+                      "Wrong type %s; strv expected",
+                      gjs_get_type_name(value));
+            return JS_FALSE;
+        }
     } else if (g_type_is_a(gtype, G_TYPE_BOXED)) {
         void *gboxed;
 
@@ -288,7 +386,7 @@ gjs_value_to_g_value(JSContext    *context,
         } else if (JSVAL_IS_OBJECT(value)) {
             JSObject *obj;
             obj = JSVAL_TO_OBJECT(value);
-            gboxed = gjs_g_boxed_from_boxed(context, obj);
+            gboxed = gjs_c_struct_from_boxed(context, obj);
         } else {
             gjs_throw(context,
                       "Wrong type %s; boxed type %s expected",
@@ -297,7 +395,10 @@ gjs_value_to_g_value(JSContext    *context,
             return JS_FALSE;
         }
 
-        g_value_set_boxed(gvalue, gboxed);
+        if (no_copy)
+            g_value_set_static_boxed(gvalue, gboxed);
+        else
+            g_value_set_boxed(gvalue, gboxed);
     } else if (g_type_is_a(gtype, G_TYPE_ENUM)) {
         if (JSVAL_IS_INT(value)) {
             GEnumValue *v;
@@ -321,9 +422,7 @@ gjs_value_to_g_value(JSContext    *context,
         }
     } else if (g_type_is_a(gtype, G_TYPE_FLAGS)) {
         if (JSVAL_IS_INT(value)) {
-            if (!_gjs_flags_value_is_valid(context,
-                                           G_FLAGS_CLASS(g_type_class_peek(gtype)),
-                                           JSVAL_TO_INT(value)))
+            if (!_gjs_flags_value_is_valid(context, gtype, JSVAL_TO_INT(value)))
                 return JS_FALSE;
 
             g_value_set_flags(gvalue, value);
@@ -396,9 +495,26 @@ gjs_value_to_g_value(JSContext    *context,
 }
 
 JSBool
-gjs_value_from_g_value(JSContext    *context,
-                       jsval        *value_p,
-                       const GValue *gvalue)
+gjs_value_to_g_value(JSContext    *context,
+                     jsval         value,
+                     GValue       *gvalue)
+{
+    return gjs_value_to_g_value_internal(context, value, gvalue, FALSE);
+}
+
+JSBool
+gjs_value_to_g_value_no_copy(JSContext    *context,
+                             jsval         value,
+                             GValue       *gvalue)
+{
+    return gjs_value_to_g_value_internal(context, value, gvalue, TRUE);
+}
+
+static JSBool
+gjs_value_from_g_value_internal(JSContext    *context,
+                                jsval        *value_p,
+                                const GValue *gvalue,
+                                gboolean      no_copy)
 {
     GType gtype;
 
@@ -455,13 +571,57 @@ gjs_value_from_g_value(JSContext    *context,
 
         obj = gjs_object_from_g_object(context, gobj);
         *value_p = OBJECT_TO_JSVAL(obj);
+    } else if (gtype == G_TYPE_STRV) {
+        if (!gjs_array_from_strv (context,
+                                  value_p,
+                                  g_value_get_boxed (gvalue))) {
+            gjs_throw(context, "Failed to convert strv to array");
+            return JS_FALSE;
+        }
+    } else if (g_type_is_a(gtype, G_TYPE_HASH_TABLE) ||
+               g_type_is_a(gtype, _array_type) ||
+               g_type_is_a(gtype, _byte_array_type) ||
+               g_type_is_a(gtype, _ptr_array_type)) {
+        gjs_throw(context,
+                  "Unable to introspect element-type of container in GValue");
+        return JS_FALSE;
     } else if (g_type_is_a(gtype, G_TYPE_BOXED)) {
+        GjsBoxedCreationFlags boxed_flags;
+        GIBaseInfo *info;
         void *gboxed;
         JSObject *obj;
 
         gboxed = g_value_get_boxed(gvalue);
+        boxed_flags = GJS_BOXED_CREATION_NONE;
 
-        obj = gjs_boxed_from_g_boxed(context, gtype, gboxed);
+        /* The only way to differentiate unions and structs is from
+         * their g-i info as both GBoxed */
+        info = g_irepository_find_by_gtype(g_irepository_get_default(),
+                                           gtype);
+        if (info == NULL) {
+            gjs_throw(context,
+                      "No introspection information found for %s",
+                      g_type_name(gtype));
+            return JS_FALSE;
+        }
+
+        switch (g_base_info_get_type(info)) {
+        case GI_INFO_TYPE_BOXED:
+        case GI_INFO_TYPE_STRUCT:
+            if (no_copy)
+                boxed_flags |= GJS_BOXED_CREATION_NO_COPY;
+            obj = gjs_boxed_from_c_struct(context, (GIStructInfo *)info, gboxed, boxed_flags);
+            break;
+        case GI_INFO_TYPE_UNION:
+            obj = gjs_union_from_c_union(context, (GIUnionInfo *)info, gboxed);
+            break;
+        default:
+            gjs_throw(context,
+                      "Unexpected introspection type %d for %s",
+                      g_base_info_get_type(info),
+                      g_type_name(gtype));
+            return JS_FALSE;
+        }
         *value_p = OBJECT_TO_JSVAL(obj);
     } else if (g_type_is_a(gtype, G_TYPE_ENUM)) {
         int v;
@@ -510,3 +670,19 @@ gjs_value_from_g_value(JSContext    *context,
 
     return JS_TRUE;
 }
+
+JSBool
+gjs_value_from_g_value(JSContext    *context,
+                       jsval        *value_p,
+                       const GValue *gvalue)
+{
+    return gjs_value_from_g_value_internal(context, value_p, gvalue, FALSE);
+}
+
+__attribute__((constructor)) static void
+_gjs_value_init_types(void)
+{
+    _array_type = g_type_from_name("GArray");
+    _byte_array_type = g_type_from_name("GByteArray");
+    _ptr_array_type = g_type_from_name("GPtrArray");
+}
