@@ -28,6 +28,8 @@
 
 G_BEGIN_DECLS
 
+#include "config.h"
+
 #if !GLIB_CHECK_VERSION(2, 18, 0)
 
 #define g_set_error_literal(error, domain, code, message) g_set_error(error, domain, code, "%s", message)
@@ -35,6 +37,30 @@ G_BEGIN_DECLS
 
 #endif
 
+/* See https://bugzilla.gnome.org/show_bug.cgi?id=622896 */
+#ifndef HAVE_JS_ADDVALUEROOT
+
+/* The old JS_AddRoot accepted anything via void *, new
+ * api is stricter.
+ */
+#define JS_AddValueRoot JS_AddRoot
+#define JS_AddObjectRoot JS_AddRoot
+#define JS_AddStringRoot JS_AddRoot
+#define JS_AddGCThingRoot JS_AddRoot
+#define JS_RemoveValueRoot JS_RemoveRoot
+#define JS_RemoveObjectRoot JS_RemoveRoot
+#define JS_RemoveStringRoot JS_RemoveRoot
+#define JS_RemoveGCThingRoot JS_RemoveRoot
+
+/* This one is complex; jsid appears to be explicitly
+ * different from JSVAL now.  If we're on an old xulrunner,
+ * define JSID_IS_VOID in a simple way.
+ */
+#define JSID_VOID JSVAL_VOID
+#define JSID_IS_VOID(id) (id == JSVAL_VOID)
+
+#endif
+
 G_END_DECLS
 
 #endif  /* __GJS_MEM_H__ */
