@@ -1,6 +1,6 @@
 /* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
 /*
- * Copyright (c) 2008  litl, LLC
+ * Copyright (c) 2008-2010  litl, LLC
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -29,6 +29,7 @@
 #include <jsapi.h>
 
 #include <gjs/gjs.h>
+#include <gjs/compat.h>
 
 #include <string.h>
 
@@ -226,7 +227,7 @@ import_native_file(JSContext  *context,
     JSBool retval = JS_FALSE;
 
     gjs_debug(GJS_DEBUG_IMPORTER,
-              "Importing '%s'", full_path);
+              "Importing '%s' from '%s'", name, full_path ? full_path : "<internal>");
 
     module_obj = JS_ConstructObject(context, NULL, NULL, NULL);
     if (module_obj == NULL) {
@@ -360,15 +361,10 @@ load_module_elements(JSContext *context,
             return;
         }
 
-        while (idp != JSVAL_VOID) {
-            jsval nameval;
+        while (!JSID_IS_VOID(idp)) {
             const char *name;
 
-            if (!JS_IdToValue(context, idp, &nameval)) {
-                continue;
-            }
-
-            if (!gjs_get_string_id(nameval, &name)) {
+            if (!gjs_get_string_id(context, idp, &name)) {
                 continue;
             }
 
@@ -509,7 +505,7 @@ do_import(JSContext  *context,
     result = JS_FALSE;
 
     filename = g_strdup_printf("%s.js", name);
-    native_filename = g_strdup_printf("%s.so", name);
+    native_filename = g_strdup_printf("%s."G_MODULE_SUFFIX, name);
     full_path = NULL;
     directories = NULL;
 
@@ -563,7 +559,15 @@ do_import(JSContext  *context,
             }
         }
 
-        /* First try importing a directory (a sub-importer) */
+        /* First try importing an internal module like byteArray */
+        if (import_native_file(context, obj, name, NULL)) {
+            gjs_debug(GJS_DEBUG_IMPORTER,
+                      "successfully imported module '%s'", name);
+            result = JS_TRUE;
+            goto out;
+        }
+
+        /* Second try importing a directory (a sub-importer) */
         if (full_path)
             g_free(full_path);
         full_path = g_build_filename(dirname, name,
@@ -591,7 +595,7 @@ do_import(JSContext  *context,
             continue;
         }
 
-        /* Second, if it's not a directory, try importing a file */
+        /* Third, if it's not a directory, try importing a file */
         g_free(full_path);
         full_path = g_build_filename(dirname, filename,
                                      NULL);
@@ -731,7 +735,7 @@ importer_new_enumerate(JSContext  *context,
             *state_p = JSVAL_NULL;
 
         if (id_p)
-            *id_p = JSVAL_ZERO;
+            *id_p = JSID_VOID;
 
         priv = priv_from_js(context, object);
         if (!priv)
@@ -820,7 +824,7 @@ importer_new_enumerate(JSContext  *context,
                 if (g_file_test(full_path, G_FILE_TEST_IS_DIR)) {
                     g_ptr_array_add(iter->elements, g_strdup(filename));
                 } else {
-                    if (g_str_has_suffix(filename, ".so") ||
+                    if (g_str_has_suffix(filename, "."G_MODULE_SUFFIX) ||
                         g_str_has_suffix(filename, ".js")) {
                         g_ptr_array_add(iter->elements,
                                         g_strndup(filename, strlen(filename) - 3));
@@ -899,7 +903,7 @@ importer_new_enumerate(JSContext  *context,
 static JSBool
 importer_new_resolve(JSContext *context,
                      JSObject  *obj,
-                     jsval      id,
+                     jsid       id,
                      uintN      flags,
                      JSObject **objp)
 {
@@ -909,7 +913,8 @@ importer_new_resolve(JSContext *context,
 
     *objp = NULL;
 
-    name = gjs_string_get_ascii(id);
+    if (!gjs_get_string_id(context, id, &name))
+        return JS_FALSE;
 
     /* let Object.prototype resolve these */
     if (strcmp(name, "valueOf") == 0 ||
@@ -925,8 +930,10 @@ importer_new_resolve(JSContext *context,
 
     /* We always import in the special load context. */
     load_context = gjs_runtime_get_load_context(JS_GetRuntime(context));
+    JS_BeginRequest(load_context);
     if (do_import(load_context, obj, priv, name)) {
         *objp = obj;
+        JS_EndRequest(load_context);
         return JS_TRUE;
     } else {
         /* Move the exception to the calling context from load context.
@@ -935,6 +942,7 @@ importer_new_resolve(JSContext *context,
             /* set an exception since none was set */
             gjs_throw(context, "No exception was set, but import failed somehow");
         }
+        JS_EndRequest(load_context);
         return JS_FALSE;
     }
 }
@@ -1182,19 +1190,25 @@ gjs_create_root_importer(JSRuntime   *runtime,
 
     context = gjs_runtime_get_load_context(runtime);
 
+    JS_BeginRequest(context);
+
     if (!gjs_object_has_property(context,
                                  JS_GetGlobalObject(context),
                                  "imports")) {
         if (gjs_define_importer(context, JS_GetGlobalObject(context),
                                 "imports",
-                                initial_search_path, add_standard_search_path) == NULL)
+                                initial_search_path, add_standard_search_path) == NULL) {
+            JS_EndRequest(context);
             return JS_FALSE;
+        }
     } else {
         gjs_debug(GJS_DEBUG_IMPORTER,
                   "Someone else already created root importer, ignoring second request");
+        JS_EndRequest(context);
         return JS_TRUE;
     }
 
+    JS_EndRequest(context);
     return JS_TRUE;
 }
 
@@ -1205,15 +1219,18 @@ gjs_define_root_importer(JSContext   *context,
 {
     JSContext *load_context;
     jsval value;
+    JSBool success;
 
+    success = JS_FALSE;
     load_context = gjs_runtime_get_load_context(JS_GetRuntime(context));
+    JS_BeginRequest(load_context);
 
     if (!gjs_object_require_property(load_context,
                                      JS_GetGlobalObject(load_context), "global object",
                                      "imports", &value) ||
         !JSVAL_IS_OBJECT(value)) {
         gjs_debug(GJS_DEBUG_IMPORTER, "Root importer did not exist, couldn't get from load context; must create it");
-        return JS_FALSE;
+        goto fail;
     }
 
     if (!JS_DefineProperty(context, in_object,
@@ -1222,8 +1239,11 @@ gjs_define_root_importer(JSContext   *context,
                            GJS_MODULE_PROP_FLAGS)) {
         gjs_debug(GJS_DEBUG_IMPORTER, "DefineProperty %s on %p failed",
                   importer_name, in_object);
-        return JS_FALSE;
+        goto fail;
     }
 
-    return JS_TRUE;
+    success = JS_TRUE;
+ fail:
+    JS_EndRequest(load_context);
+    return success;
 }
