@@ -26,6 +26,7 @@
 #include "dbus-values.h"
 
 #include "../gjs/gjs.h"
+#include "../gjs/compat.h"
 #include "../gi/closure.h"
 
 #include <util/log.h>
@@ -36,11 +37,11 @@ static DBusConnection *session_bus = NULL;
 static gboolean system_bus_weakref_added = FALSE;
 static DBusConnection *system_bus = NULL;
 
-/* A global not-threadsafe reference to the DBusMessage
+/* A global not-threadsafe stack of DBusMessages
  * currently being processed when invoking a user callback.
  * Accessible via dbus.currentMessageContext.
  */
-static DBusMessage *_gjs_current_dbus_message = NULL;
+static GSList *_gjs_current_dbus_messages = NULL;
 
 #define DBUS_CONNECTION_FROM_TYPE(type) ((type) == DBUS_BUS_SESSION ? session_bus : system_bus)
 
@@ -95,14 +96,21 @@ bus_check(JSContext *context, DBusBusType bus_type)
 }
 
 void
-gjs_js_set_current_message (DBusMessage        *message)
+gjs_js_push_current_message(DBusMessage *message)
 {
-  g_assert ((message == NULL && _gjs_current_dbus_message != NULL) ||
-            (message != NULL && _gjs_current_dbus_message == NULL));
   /* Don't need to take a ref here, if someone forgets to unset the message,
    * that's a bug.
    */
-  _gjs_current_dbus_message = message;
+  _gjs_current_dbus_messages = g_slist_prepend(_gjs_current_dbus_messages, message);
+}
+
+void
+gjs_js_pop_current_message(void)
+{
+    g_assert(_gjs_current_dbus_messages != NULL);
+
+    _gjs_current_dbus_messages = g_slist_remove_link(_gjs_current_dbus_messages,
+                                                     _gjs_current_dbus_messages);
 }
 
 static DBusMessage*
@@ -245,7 +253,10 @@ complete_call(JSContext   *context,
     g_assert(ret_values != NULL);
 
     array_length = gjs_rooted_array_get_length(context, ret_values);
-    if (array_length == 1) {
+    if (array_length == 0) {
+        /* the callback expects to be called with callback(undefined, null) */
+        *retval = JSVAL_VOID;
+    } else if (array_length == 1) {
         /* If the array only has one element return that element alone */
         *retval = gjs_rooted_array_get(context,
                                           ret_values,
@@ -303,9 +314,9 @@ pending_notify(DBusPendingCall *pending,
     /* argv[0] will be the return value if any, argv[1] we fill with exception if any */
     gjs_set_values(context, argv, 2, JSVAL_NULL);
     gjs_root_value_locations(context, argv, 2);
-    gjs_js_set_current_message(reply);
+    gjs_js_push_current_message(reply);
     complete_call(context, &argv[0], reply, &derror);
-    gjs_js_set_current_message(NULL);
+    gjs_js_pop_current_message();
     g_assert(!dbus_error_is_set(&derror)); /* not supposed to be left set by complete_call() */
 
     /* If completing the call failed, we still call the callback, but with null for the reply
@@ -316,9 +327,9 @@ pending_notify(DBusPendingCall *pending,
         JS_ClearPendingException(context);
     }
 
-    gjs_js_set_current_message(reply);
+    gjs_js_push_current_message(reply);
     gjs_closure_invoke(closure, 2, &argv[0], &discard);
-    gjs_js_set_current_message(NULL);
+    gjs_js_pop_current_message();
 
     if (reply)
         dbus_message_unref(reply);
@@ -632,9 +643,12 @@ signal_handler_callback(DBusConnection *connection,
         return;
     }
 
+    JS_BeginRequest(context);
+
     dbus_message_iter_init(message, &arg_iter);
     if (!gjs_js_values_from_dbus(context, &arg_iter, &arguments)) {
         gjs_debug(GJS_DEBUG_DBUS, "Failed to marshal dbus signal to JS");
+        JS_EndRequest(context);
         return;
     }
 
@@ -647,16 +661,19 @@ signal_handler_callback(DBusConnection *connection,
     gjs_debug(GJS_DEBUG_DBUS,
               "Invoking closure on signal received, %d args",
               gjs_rooted_array_get_length(context, arguments));
-    gjs_js_set_current_message(message);
+
+    gjs_js_push_current_message(message);
     gjs_closure_invoke(handler->closure,
                        gjs_rooted_array_get_length(context, arguments),
                        gjs_rooted_array_get_data(context, arguments),
                        &ret_val);
-    gjs_js_set_current_message(NULL);
+    gjs_js_pop_current_message();
 
     gjs_rooted_array_free(context, arguments, TRUE);
 
     signal_handler_unref(handler); /* for safety */
+
+    JS_EndRequest(context);
 }
 
 /* Args are bus_name, object_path, iface, signal, and callback */
@@ -997,19 +1014,23 @@ on_name_acquired(DBusConnection *connection,
         return;
     }
 
+    JS_BeginRequest(context);
+
     argc = 1;
 
     argv[0] = STRING_TO_JSVAL(JS_NewStringCopyZ(context, name));
-    JS_AddRoot(context, &argv[0]);
+    JS_AddValueRoot(context, &argv[0]);
 
     rval = JSVAL_VOID;
-    JS_AddRoot(context, &rval);
+    JS_AddValueRoot(context, &rval);
 
     gjs_closure_invoke(owner->acquired_closure,
                        argc, argv, &rval);
 
-    JS_RemoveRoot(context, &argv[0]);
-    JS_RemoveRoot(context, &rval);
+    JS_RemoveValueRoot(context, &argv[0]);
+    JS_RemoveValueRoot(context, &rval);
+
+    JS_EndRequest(context);
 }
 
 static void
@@ -1032,19 +1053,23 @@ on_name_lost(DBusConnection *connection,
         return;
     }
 
+    JS_BeginRequest(context);
+
     argc = 1;
 
     argv[0] = STRING_TO_JSVAL(JS_NewStringCopyZ(context, name));
-    JS_AddRoot(context, &argv[0]);
+    JS_AddValueRoot(context, &argv[0]);
 
     rval = JSVAL_VOID;
-    JS_AddRoot(context, &rval);
+    JS_AddValueRoot(context, &rval);
 
     gjs_closure_invoke(owner->lost_closure,
                           argc, argv, &rval);
 
-    JS_RemoveRoot(context, &argv[0]);
-    JS_RemoveRoot(context, &rval);
+    JS_RemoveValueRoot(context, &argv[0]);
+    JS_RemoveValueRoot(context, &rval);
+
+    JS_EndRequest(context);
 }
 
 static void
@@ -1209,6 +1234,8 @@ on_name_appeared(DBusConnection *connection,
         return;
     }
 
+    JS_BeginRequest(context);
+
     argc = 2;
 
     gjs_set_values(context, argv, argc, JSVAL_VOID);
@@ -1218,13 +1245,15 @@ on_name_appeared(DBusConnection *connection,
     argv[1] = STRING_TO_JSVAL(JS_NewStringCopyZ(context, owner_unique_name));
 
     rval = JSVAL_VOID;
-    JS_AddRoot(context, &rval);
+    JS_AddValueRoot(context, &rval);
 
     gjs_closure_invoke(watcher->appeared_closure,
                           argc, argv, &rval);
 
-    JS_RemoveRoot(context, &rval);
+    JS_RemoveValueRoot(context, &rval);
     gjs_unroot_value_locations(context, argv, argc);
+
+    JS_EndRequest(context);
 }
 
 static void
@@ -1248,6 +1277,8 @@ on_name_vanished(DBusConnection *connection,
         return;
     }
 
+    JS_BeginRequest(context);
+
     argc = 2;
 
     gjs_set_values(context, argv, argc, JSVAL_VOID);
@@ -1257,13 +1288,15 @@ on_name_vanished(DBusConnection *connection,
     argv[1] = STRING_TO_JSVAL(JS_NewStringCopyZ(context, owner_unique_name));
 
     rval = JSVAL_VOID;
-    JS_AddRoot(context, &rval);
+    JS_AddValueRoot(context, &rval);
 
     gjs_closure_invoke(watcher->vanished_closure,
                           argc, argv, &rval);
 
-    JS_RemoveRoot(context, &rval);
+    JS_RemoveValueRoot(context, &rval);
     gjs_unroot_value_locations(context, argv, argc);
+
+    JS_EndRequest(context);
 }
 
 static const GjsDBusWatchNameFuncs watch_name_funcs = {
@@ -1380,7 +1413,7 @@ gjs_js_dbus_watch_name(JSContext  *context,
 static JSBool
 unique_name_getter(JSContext  *context,
                    JSObject   *obj,
-                   jsval       id,
+                   jsid        id,
                    jsval      *value_p)
 {
     const char *name;
@@ -1390,7 +1423,8 @@ unique_name_getter(JSContext  *context,
     if (!get_bus_type_from_object(context, obj, &bus_type))
         return JS_FALSE;
 
-    name = gjs_string_get_ascii(id);
+    if (!gjs_get_string_id(context, id, &name))
+        return JS_FALSE;
 
     gjs_debug_jsprop(GJS_DEBUG_DBUS, "Get prop '%s' on dbus object", name);
 
@@ -1490,7 +1524,7 @@ gjs_js_dbus_start_service(JSContext  *context,
 static JSBool
 gjs_js_dbus_get_machine_id(JSContext *context,
                            JSObject  *obj,
-                           jsval      key,
+                           jsid       key,
                            jsval     *value)
 {
     char *machine_id;
@@ -1524,23 +1558,26 @@ gjs_js_dbus_get_current_message_context(JSContext  *context,
     JSObject *context_obj;
     jsval context_val;
     JSBool result = JS_FALSE;
+    DBusMessage *current_message;
 
     if (!gjs_parse_args(context, "getCurrentMessageContext", "", argc, argv))
         return JS_FALSE;
 
-    if (!_gjs_current_dbus_message) {
+    if (!_gjs_current_dbus_messages) {
         *retval = JSVAL_NULL;
         return JS_TRUE;
     }
 
+    current_message = _gjs_current_dbus_messages->data;
+
     context_obj = JS_ConstructObject(context, NULL, NULL, NULL);
     if (context_obj == NULL)
         return JS_FALSE;
 
     context_val = OBJECT_TO_JSVAL(context_obj);
-    JS_AddRoot(context, &context_val);
+    JS_AddValueRoot(context, &context_val);
 
-    sender = dbus_message_get_sender(_gjs_current_dbus_message);
+    sender = dbus_message_get_sender(current_message);
     if (sender)
         sender_str = JS_NewStringCopyZ(context, sender);
     else
@@ -1554,7 +1591,7 @@ gjs_js_dbus_get_current_message_context(JSContext  *context,
         goto out;
 
     if (!JS_DefineProperty(context, context_obj,
-                           "serial", INT_TO_JSVAL(dbus_message_get_serial(_gjs_current_dbus_message)),
+                           "serial", INT_TO_JSVAL(dbus_message_get_serial(current_message)),
                            NULL, NULL,
                            JSPROP_ENUMERATE))
         goto out;
@@ -1563,7 +1600,7 @@ gjs_js_dbus_get_current_message_context(JSContext  *context,
     *retval = context_val;
 
 out:
-    JS_RemoveRoot(context, &context_val);
+    JS_RemoveValueRoot(context, &context_val);
     return result;
 }
 
@@ -1580,7 +1617,7 @@ define_bus_proto(JSContext *context,
     result = JS_FALSE;
 
     bus_proto_val = JSVAL_VOID;
-    JS_AddRoot(context, &bus_proto_val);
+    JS_AddValueRoot(context, &bus_proto_val);
 
     bus_proto_obj = JS_ConstructObject(context, NULL, NULL, NULL);
     if (bus_proto_obj == NULL)
@@ -1674,7 +1711,7 @@ define_bus_proto(JSContext *context,
     result = JS_TRUE;
 
  out:
-    JS_RemoveRoot(context, &bus_proto_val);
+    JS_RemoveValueRoot(context, &bus_proto_val);
 
     return result;
 }
@@ -1699,7 +1736,7 @@ define_bus_object(JSContext  *context,
     result = JS_FALSE;
 
     bus_val = JSVAL_VOID;
-    JS_AddRoot(context, &bus_val);
+    JS_AddValueRoot(context, &bus_val);
 
     bus_obj = JS_ConstructObject(context, NULL, proto_obj, NULL);
     if (bus_obj == NULL)
@@ -1735,7 +1772,7 @@ define_bus_object(JSContext  *context,
     result = JS_TRUE;
 
  out:
-    JS_RemoveRoot(context, &bus_val);
+    JS_RemoveValueRoot(context, &bus_val);
 
     return result;
 }
